package checkers.jcip;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import javax.lang.model.element.Element;
import javax.lang.model.element.ExecutableElement;
import javax.lang.model.element.Modifier;

import com.sun.source.tree.*;

import checkers.basetype.BaseTypeChecker;
import checkers.basetype.BaseTypeVisitor;
import checkers.jcip.quals.GuardedBy;
import checkers.jcip.quals.Holding;
import checkers.nullness.NullnessChecker;
import checkers.source.Result;
import checkers.util.TreeUtils;

/**
 * A type-checking visitor for the JCIP concurrency type system.
 * The visitor reports errors or warnings for violations for the
 * following cases:
 *
 * <ol>
 * <li value="1">A guarded field is dereferenced without holding its lock
 * <li value="2">A guarded non-final non-volatile field is accessed
 *      (merely referenced) without holding its lock
 * <p>TODO: Doesn't make sense to separate dereferencing and accessing?
 * <li value="3">A method is invoked without the caller holding the locks
 *      it requires
 * </ol>
 *
 * Self-accesses (via {@code this} or {@code super}) can always be dereferenced.
 *
 * @see NullnessChecker
 */
public class JCIPVisitor extends BaseTypeVisitor<Void, Void> {

    private List<String> locksHolding;
    public JCIPVisitor(BaseTypeChecker checker, CompilationUnitTree root) {
        super(checker, root);
        locksHolding = new ArrayList<String>();
    }


    public Void visitMethod(MethodTree node, Void p) {
        List<String> prevHolding = new ArrayList<String>(locksHolding);
        ExecutableElement method = TreeUtils.elementFromDeclaration(node);
        Holding holdingAnno = method.getAnnotation(Holding.class);
        if (holdingAnno != null) {
            locksHolding.addAll(Arrays.asList(holdingAnno.value()));
        }

        if (method.getModifiers().contains(Modifier.SYNCHRONIZED)) {
            locksHolding.add("this");
        }

        try {
            return super.visitMethod(node, p);
        } finally {
            locksHolding = prevHolding;
        }
    }

    public Void visitSynchronized(SynchronizedTree node, Void p) {
        List<String> prevHolding = new ArrayList<String>(locksHolding);

        prevHolding.add(node.getExpression().toString());

        try {
            return super.visitSynchronized(node, p);
        } finally {
            locksHolding = prevHolding;
        }
    }

    public Void visitIdentifier(IdentifierTree node, Void p) {
        Element elem = TreeUtils.elementFromUse(node);
        if (elem.getKind().isField()) {
            GuardedBy guardedByAnno = elem.getAnnotation(GuardedBy.class);
            if (guardedByAnno != null) {
                if (!locksHolding.contains(guardedByAnno.value())) {
                    checker.report(Result.failure("unguarded.access"), node);
                }
            }
        }
        
        return super.visitIdentifier(node, p);
    }
}
